/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.corbaservices;

//---IMPORTS---
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * MiniWebServer
 *
 * This provides a very basic webserver for serving IOR files.
 * This could easily serve other text files, but not binaries.
 *
 * @author  $Author: tdb $
 * @version $Id: MiniWebServer.java,v 1.9 2002/05/21 16:47:11 tdb Exp $
 */
public class MiniWebServer extends Thread {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public static final String REVISION = "$Revision: 1.9 $";
    
//---STATIC METHODS---

//---CONSTRUCTORS---
    
    /**
     * Construct a new MiniWebServer
     *
     * @param port The port to bind to
     * @param okUrlsConfig The URL list config file
     * @param webDir The web directory
     */
    public MiniWebServer(int port, String okUrlsConfig, String webDir) {
        _port = port;
        _okUrlsConfig = okUrlsConfig;
        _webDir = webDir;
    }

//---PUBLIC METHODS---

    /**
     * Main thread of execution. Keeps looping spawning off
     * Handler's to deal with each client. This should keep
     * going, and extra care has been taking with error
     * catching.
     */
    public void run() {
        // carry on forever
        while(true) {
            boolean run = true;
            ServerSocket serverSocket = null;
            try {
                serverSocket = new ServerSocket(_port);
            } catch (IOException e) {
                run = false;
                System.out.println(e);
                e.printStackTrace();
            }
            while(run) {
                try {
                    // get a new client and give it to a Handler
                    Socket socket = serverSocket.accept();
                    Handler handler = new Handler(socket);
                    handler.start();
                } catch (IOException e) {
                    run=false;
                    System.out.println(e);
                    e.printStackTrace();
                }
            }
        }
    }

//---PRIVATE METHODS---

    /**
     * Reads a file in given an URL. Very basic, and has only
     * been tested to work for URL's in the / directory.
     *
     * @param url The URL given to the webserver
     * @return A String containing data from the file
     */
    private String readUrl(String url) {
        // trim off the initial /
        if(url.startsWith("/")) {
            url = url.substring(1);
        }
        try {
            // need to prefix with web directory
            File file = new File(_webDir+"/"+url);
            if(file.exists() && file.canRead()) {
                // read in all the data
                String data = "";
                BufferedReader reader = new BufferedReader(new FileReader(file));
                while(reader.ready()) {
                    String line = reader.readLine();
                    if(line != null) {
                        data += line + "\n";
                    }
                }
                return data;
            }
            else {
                // return null if we can't read it
                return null;
            }
        } catch (IOException e) {
            System.out.println(e);
            e.printStackTrace();
            return null;
        }
    }
    
    /**
     * Checks if an URL is allowed. Failure of this method
     * indicates a client should be sent a 403.
     *
     * @param url the url to check
     * @return whether the url is permitted
     */
    private boolean okUrl(String url) {
        // recheck the list - this means the list can be
        // changed on the fly :)
        readOkUrls();
        Iterator i = _okUrls.iterator();
        while(i.hasNext()) {
            if(url.equals((String)i.next())) {
                // matched url against our list
                return true;
            }
        }
        return false;
    }
    
    /**
     * Reads the list of URLs into a list.
     */
    private void readOkUrls() {
        File file = new File(_okUrlsConfig);
        long lastMod = file.lastModified();
        // only want to re-read the list if it's changed
        if(lastMod > _okUrlsConfigStamp) {
            _okUrlsConfigStamp = lastMod;
            _okUrls = new LinkedList();
            if(file.exists() && file.canRead()) {
                try {
                    BufferedReader reader = new BufferedReader(new FileReader(file));
                    while(reader.ready()) {
                        String line = reader.readLine();
                        // ignore comments
                        if(!line.startsWith("#")) {
                            _okUrls.add(line);
                        }
                    }
                } catch(IOException e) {
                    System.out.println(e);
                    e.printStackTrace();
                }
            }
        }
    }
    
//---ACCESSOR/MUTATOR METHODS---

//---ATTRIBUTES---
    
    /**
     * The port we're bound to.
     */
    private int _port;
    
    /**
     * List of URLs that are permitted
     */
    private LinkedList _okUrls;
    
    /**
     * Config file for the URL list
     */
    private String _okUrlsConfig;
    
    /**
     * Timestamp of the URL config file
     */
    private long _okUrlsConfigStamp = 0;
    
    /**
     * Web directory
     */
    private String _webDir;

//---STATIC ATTRIBUTES---

//---INNER CLASSES---
    
    /**
     * A Handler to deal with each client. This is an inner
     * class because I felt like trying it out :)
     */
    private class Handler extends Thread {
        
        /**
         * Construct a new Handler
         *
         * @param socket The socket connected to the client
         */
        public Handler(Socket socket) {
            _socket = socket;
        }
        
        /**
         * Start the Handler running. This will read the URL
         * from the client, and respond accordinly.
         *
         * This does not try to deal with bad data
         */
        public void run() {
            BufferedReader reader = null;
            PrintWriter writer = null;
            try {
                // setup the reader and writer
                reader = new BufferedReader(new InputStreamReader(_socket.getInputStream()));
                writer = new PrintWriter(_socket.getOutputStream(), true);
                String line= "";
                String url = "";
                // keep reading from the client until it sends no more...
                do {
                    line = reader.readLine();
                    // ah, this contains the URL
                    if(line.startsWith("GET ")) {
                        // strip off "GET "
                        url = line.substring(4);
                        // strip off anything past the URL
                        if(url.lastIndexOf(' ') != -1) {
                            url = url.substring(0, url.lastIndexOf(' '));
                        }
                    }
                } while(!line.equals(""));
                // check the URL is allowed
                if(okUrl(url)) {
                    // read the file
                    String data = readUrl(url);
                    // check the file could be read, null means it couldn't
                    if(data != null) {
                        writer.println(data);
                    }
                    else {
                        // oh dear, couldn't find it, send a 404
                        writer.println("HTTP/1.1 404 Not Found\n");
                        writer.println("404 Not Found");
                    }
                }
                else {
                    // oh dear, not allowed, send a 403
                    writer.println("HTTP/1.1 403 Permission Denied\n");
                    writer.println("403 Permission Denied");
                }
            } catch(IOException e) {
                System.out.println(e);
                e.printStackTrace();
            }
            // clean up and close
            try {
                reader.close();
                writer.close();
                _socket.close();
            } catch(IOException e) {
                System.out.println(e);
                e.printStackTrace();
            }
        }
        
        /**
         * The socket connected to the client
         */
        private Socket _socket;
    }
    
}
